@@ -9,27 +9,24 @@ from django.shortcuts import redirect
 from django.conf import settings
 from django.template.defaultfilters import slugify
 from django.utils.datastructures import SortedDict
-from viewsets.views import AutocompleteListView
+from viewsets.views import MiningListView
+from copy import copy, deepcopy
+import six
 
 
 class ViewSetMixin(object):
 
-    def dispatch(self, request, *args, **kwargs):
-        if hasattr(self, "manager"):
-            retval = self.manager.pre_dispatch(request, self, *args, **kwargs)
-            if retval:
-                return retval
-        return super(ViewSetMixin, self).dispatch(request, *args, **kwargs)
-
     def get_context_data(self, **kwargs):
         context = super(ViewSetMixin, self).get_context_data(**kwargs)
-        context.update(self.manager.extra_context(self.request, self))
+        if getattr(self, "manager", None):
+            context.update(self.manager.extra_context(self.request, self))
         return context
 
     # need to set the current app for url namespace resolution
     def render_to_response(self, context, **response_kwargs):
-        response_kwargs["current_app"] = self.manager.name
-        context.update({"current_app": self.manager.name})
+        if getattr(self, "manager", None):
+            response_kwargs["current_app"] = self.manager.name
+            context.update({"current_app": self.manager.name})
 
         return super(ViewSetMixin, self).render_to_response(context, **response_kwargs)
 
@@ -38,40 +35,30 @@ class ViewSetMixin(object):
             current_app=self.manager.name)
 
     def get_template_names(self):
-        template = getattr(self, "template", None)
-        if template:
-            return template
+        template_name = getattr(self, "template_name", None)
+        if template_name:
+            return template_name
+
+        templates = [
+            [
+                self.manager.base_template_dir,
+                self.manager.template_dir,
+                self.name + ".html"
+            ],
+            [
+                self.manager.base_template_dir,
+                self.manager.default_app,
+                self.name + ".html"
+            ]
+        ]
 
         if self.request.is_ajax():
-            ajax = "_ajax"
-        else:
-            ajax = ""
+            ajax_templates = deepcopy(templates)
+            for template in ajax_templates:
+                template[-1] = self.name + "_ajax.html"
+            templates = ajax_templates + templates
 
-        if self.manager.base_template_dir:
-            templates = [
-                "%s/%s/%s%s.html" % (
-                    self.manager.base_template_dir,
-                    self.manager.template_dir,
-                    self.name,
-                    ajax),
-                "%s/%s/%s%s.html" % (
-                    self.manager.base_template_dir,
-                    self.manager.default_app,
-                    self.name,
-                    ajax)
-            ]
-        else:
-            templates = [
-                "%s/%s%s.html" % (
-                    self.manager.template_dir,
-                    self.name,
-                    ajax),
-                "%s/%s%s.html" % (
-                    self.manager.default_app,
-                    self.name,
-                    ajax)
-            ]
-        return templates
+        return [os.path.join(*bits) for bits in templates]
 
     def get_queryset(self):
         return self.manager.get_queryset(self, self.request, **self.kwargs)
@@ -92,7 +79,7 @@ class ViewSetDeleteView(ViewSetMixin, DeleteView):
             current_app=self.manager.name)
 
 
-class ViewSetListView(ViewSetMixin, AutocompleteListView):
+class ViewSetListView(ViewSetMixin, MiningListView):
     paginate_by = 25
 
 
@@ -100,25 +87,38 @@ class ViewSetDetailView(ViewSetMixin, DetailView):
     pass
 
 
+class classproperty(property):
+    def __get__(self, cls, owner):
+        return self.fget.__get__(None, owner)()
+
+
 class ViewSet(object):
     PK_URL = "(?P<pk>\d+)/"
     SLUG_URL = "(?P<slug>[\w\-\d]+)/"
 
+    mixin = None
     name = None
     model = None
+    opts = None
     base_template_dir = ""
     base_url = None
     object_url = PK_URL
     template_dir = None
     default_app = "base"
-    managers = []
+    _managers = []
     exclude = []
 
-    def __init__(self, name=None, model=None, template_dir=None, exclude=[]):
+    def __init__(self, name=None, model=None, template_dir=None, exclude=None):
+
+        if exclude:
+            self.exclude = exclude
 
         if model:
             self.model = model
 
+        if self.model:
+            self.opts = self.model._meta
+
         if name:
             self.name = name
         elif not self.name:
@@ -137,8 +137,7 @@ class ViewSet(object):
             ("list", (ViewSetListView, r'^$')),
         ))
 
-        exclude.extend(self.exclude)
-        for name in exclude:
+        for name in self.exclude:
             del self.views[name]
 
         if template_dir:
@@ -146,17 +145,14 @@ class ViewSet(object):
         elif not self.template_dir:
             self.template_dir = self.name
 
-        ViewSet.managers.append(self)
+        ViewSet._managers.append(self)
 
-    def protect(self, function):
-        if not hasattr(function, "protected"):
-            def inner(self, request, *args, **kwargs):
-                if not request.user.is_authenticated():
-                    return redirect("%s?next=%s" % (settings.LOGIN_URL, request.path))
-                return function(self, request, *args, **kwargs)
-            inner.protected = True
-            return inner
-        return function
+    @classproperty
+    @classmethod
+    def managers(klass):
+        for m in klass._managers:
+            if isinstance(m, klass):
+                yield m
 
     def pre_dispatch(self, request, view, **kwargs):
         pass
@@ -168,17 +164,34 @@ class ViewSet(object):
 
             # override here or these views can't be used elsewhere
             # they will be slightly modified soon
-            class view(view_class):
-                pass
+            if self.mixin and not issubclass(view_class, self.mixin):
+                parents = (self.mixin, view_class)
+            else:
+                parents = (view_class,)
 
-            view.dispatch = self.protect(view.dispatch)
+            view = type(
+                "%s_%s_%s" % (
+                    self.__class__.__name__,
+                    self.model._meta.object_name,
+                    name
+                ),
+                parents,
+                {
+                    "name": name,
+                    "model": self.model,
+                    "manager": self
+                })
 
             # preserve csrf setting
             if getattr(view_class.dispatch, "csrf_exempt", False):
-                view.dispatch.__func__.csrf_exempt = True
+                if six.PY3:
+                    view.dispatch.csrf_exempt = True
+                else:
+                    view.dispatch.__func__.csrf_exempt = True
+                    
 
-            view.name = name
-            view.manager = self
+#             view.name = name
+#             view.manager = self
             view = view.as_view(**kwargs)
             urls.append((url_regex, view, {}, name))
 
@@ -186,7 +199,9 @@ class ViewSet(object):
 
     @classmethod
     def all_urls(klass):
-        urls = [(m.base_url, include(m.get_urls())) for m in klass.managers]
+        urls = []
+        for m in klass.managers:
+            urls.append((m.base_url, include(m.get_urls())))
         return patterns('', *urls)
 
     def get_queryset(self, view, request, **kwargs):
@@ -214,4 +229,8 @@ class ViewSet(object):
             r'^%s%s/$' % (self.object_url, name))
 
     def extra_context(self, request, view, **kwargs):
-        return dict(manager=self, opts=self.model._meta, **kwargs)
+        return dict(
+            manager=self,
+            opts=self.model._meta,
+            **kwargs
+        )
